#include "CMqttStandardThread.h"

#include "NetworkService/CModemSetupThread.h"

#define CONFIG_MQTT_BROKER_HOSTNAME "mqtt.nordicsemi.academy"

sockaddr_storage CMqttStandardThread::broker = {};

mqtt_client CMqttStandardThread::client = {};

CMqttStandardThread *CMqttStandardThread::instance = nullptr;

// const char* CMqttStandardThread::MQTT_TOPIC_COMANDI_REMOTI = "/cmds";

// const char* CMqttStandardThread::MQTT_TOPIC_NEW_RELEASE = "/verde";

// const char* CMqttStandardThread::MQTT_TOPIC_TEST_OK = "/blu";


uint8_t CMqttStandardThread::rx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE] ={};

uint8_t CMqttStandardThread::tx_buffer[CONFIG_MQTT_MESSAGE_BUFFER_SIZE] ={};

uint8_t CMqttStandardThread::payload_buf[CONFIG_MQTT_PAYLOAD_BUFFER_SIZE] ={};



CMqttStandardThread::CMqttStandardThread():CBaseMqtt()
{

	instance = this;

	initTopicSubscription();

	createTimer();
	
}

CMqttStandardThread::~CMqttStandardThread()
{

}

void CMqttStandardThread::initTopicSubscription(){

    sprintf((char*)topicSubscribed[0], "%s%s", CBaseThread::SERIAL_NUMBER, MQTT_TOPIC_COMANDI_REMOTI);

    sprintf((char*)topicSubscribed[1], "%s%s", CBaseThread::SERIAL_NUMBER, MQTT_TOPIC_NEW_RELEASE);

    sprintf((char*)topicSubscribed[2], "%s%s", CBaseThread::SERIAL_NUMBER, MQTT_TOPIC_TEST_OK);

    subscribe_topics[0].topic.utf8 = (uint8_t*)topicSubscribed[0];
    subscribe_topics[0].topic.size = strlen((const char*)topicSubscribed[0]);
    subscribe_topics[0].qos = MQTT_QOS_1_AT_LEAST_ONCE;

    subscribe_topics[1].topic.utf8 = (uint8_t*)topicSubscribed[1];
    subscribe_topics[1].topic.size = strlen((const char*)topicSubscribed[1]);
    subscribe_topics[1].qos = MQTT_QOS_1_AT_LEAST_ONCE;

    subscribe_topics[2].topic.utf8 = (uint8_t*)topicSubscribed[2];
    subscribe_topics[2].topic.size = strlen((const char*)topicSubscribed[2]);
    subscribe_topics[2].qos = MQTT_QOS_1_AT_LEAST_ONCE;
	
    subscription_list.list = subscribe_topics;
    subscription_list.list_count = ARRAY_SIZE(subscribe_topics);
    subscription_list.message_id = 1234;

}

void CMqttStandardThread::onTimerCallback(){


    CLogger::getInstance()->log("Timeout connecting to MQTT broker\n");

    status = MQTT_BROKER_STATE_DISCONNECTING;

}


int CMqttStandardThread::broker_init(void)
{
	int err;
	struct addrinfo *result;
	struct addrinfo *addr;
	struct addrinfo hints = {
		.ai_family = AF_INET,
		.ai_socktype = SOCK_STREAM
	};

	err = getaddrinfo(CONFIG_MQTT_BROKER_HOSTNAME, NULL, &hints, &result);
	if (err) {
		CLogger::getInstance()->log("getaddrinfo failed: %d", err);
		return -ECHILD;
	}

	addr = result;

	/* Look for address of the broker. */
	while (addr != NULL) {
		/* IPv4 Address. */
		if (addr->ai_addrlen == sizeof(struct sockaddr_in)) {
			struct sockaddr_in *broker4 =
				((struct sockaddr_in *)&broker);
			char ipv4_addr[NET_IPV4_ADDR_LEN];

			broker4->sin_addr.s_addr =
				((struct sockaddr_in *)addr->ai_addr)
				->sin_addr.s_addr;
			broker4->sin_family = AF_INET;
			broker4->sin_port = htons(CONFIG_MQTT_HELPER_PORT);

			inet_ntop(AF_INET, &broker4->sin_addr.s_addr,
				  ipv4_addr, sizeof(ipv4_addr));
				  CLogger::getInstance()->log("IPv4 Address found %s", (char *)(ipv4_addr));

			break;
		} else {
			CLogger::getInstance()->log("ai_addrlen = %u should be %u or %u",
				(unsigned int)addr->ai_addrlen,
				(unsigned int)sizeof(struct sockaddr_in),
				(unsigned int)sizeof(struct sockaddr_in6));
		}

		addr = addr->ai_next;
	}

	/* Free the address. */
	freeaddrinfo(result);

	return err;
}



int CMqttStandardThread::client_init(struct mqtt_client *client)
{
	int err = 0;
	/* Initializes the client instance. */
	mqtt_client_init(client);

	/* Resolves the configured hostname and initializes the MQTT broker structure */
	err = broker_init();
	if (err) {
		CLogger::getInstance()->log("Failed to initialize broker connection");
		return err;
	}

	/* MQTT client configuration */
	client->broker = &broker;
	client->evt_cb = &CMqttStandardThread::mqtt_evt_handler;
	client->client_id.utf8 = CModemSetupThread::client_id_get();

	client->client_id.size = strlen(client->client_id.utf8);
	client->password = NULL;
	client->user_name = NULL;
	client->protocol_version = MQTT_VERSION_3_1_1;

	/* MQTT buffers configuration */
	client->rx_buf = rx_buffer;
	client->rx_buf_size = sizeof(rx_buffer);
	client->tx_buf = tx_buffer;
	client->tx_buf_size = sizeof(tx_buffer);

	/* STEP 5 - Modify the client client_init() function to use Secure TCP transport instead of non-secure TCP transport.  */

	//get a pointer to the TLS configration 
	struct mqtt_sec_config *tls_cfg = &(client->transport).tls.config;

	static sec_tag_t sec_tag_list[] = { CONFIG_MQTT_HELPER_SEC_TAG  };

	CLogger::getInstance()->log("TLS enabled");
	client->transport.type = MQTT_TRANSPORT_SECURE;

	tls_cfg->peer_verify = TLS_PEER_VERIFY_REQUIRED;
	tls_cfg->cipher_list = NULL;
	tls_cfg->cipher_count = 0;
	tls_cfg->sec_tag_count = ARRAY_SIZE(sec_tag_list);
	tls_cfg->sec_tag_list = sec_tag_list;
	tls_cfg->hostname = CONFIG_MQTT_BROKER_HOSTNAME;

	tls_cfg->session_cache = IS_ENABLED(CONFIG_MQTT_TLS_SESSION_CACHING) ?
					    TLS_SESSION_CACHE_ENABLED :
					    TLS_SESSION_CACHE_DISABLED;


	return err;
}

void CMqttStandardThread::mqtt_evt_handler(struct mqtt_client *const c,
	const struct mqtt_evt *evt)
{

		switch (evt->type) {
			case MQTT_EVT_CONNACK:
				if (evt->result != 0) {

					instance->status = MQTT_BROKER_STATE_DISCONNECTED;
					CLogger::getInstance()->log("MQTT connect failed: %d", evt->result);
					break;
				}

				instance->status = MQTT_BROKER_STATE_CONNECTED;

				instance->stopOneShotTimer();
		
				mqtt_subscribe(c, &instance->subscription_list);

				CLogger::getInstance()->log("MQTT client connected");


				break;
		
			case MQTT_EVT_DISCONNECT:

				instance->status = MQTT_BROKER_STATE_DISCONNECTED;
				CLogger::getInstance()->log("MQTT client disconnected: %d", evt->result);
				break;

			case MQTT_EVT_PUBLISH:

				break;
			default:
				break;
				
			}

	
}

void CMqttStandardThread::runHandler()
{


	int err = 0;
	CLogger::getInstance()->log("MQTT STANDARD Thread started\n");


	while(true)
	{

        uint32_t events = k_event_wait(&CBaseThread::lte_event_flags, LTE_CONNECTED_FLAG, false, K_FOREVER);

        if (events & LTE_CONNECTED_FLAG) {
			if (instance->status == MQTT_BROKER_STATE_DISCONNECTED)
			{

				client_init(&client);

				err = mqtt_connect(&client);

				if (err) {
					
					CLogger::getInstance()->log("Error in mqtt_connect: %d", err);
		
				}

				startOneShotTimer(20000);

				instance->status = MQTT_BROKER_STATE_CONNECTING;

			}
			else if  (MQTT_BROKER_STATE_DISCONNECTING == status){

				mqtt_disconnect(&client);

				instance->status = MQTT_BROKER_STATE_DISCONNECTED;
			}
			else if  (MQTT_BROKER_STATE_CONNECTED == status)
			{


			}

		}


		k_sleep(K_SECONDS(1));
	}


}